=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2009-05-07 05:08:46 +0000
+++ bzrlib/builtins.py	2009-05-11 18:35:20 +0000
@@ -4866,13 +4866,7 @@
     def _run(self, submit_branch, revision, public_branch, remember, format,
              no_bundle, no_patch, output, from_, mail_to, message, body):
         from bzrlib.revision import NULL_REVISION
-        branch = Branch.open_containing(from_)[0]
-        if output is None:
-            outfile = cStringIO.StringIO()
-        elif output == '-':
-            outfile = self.outf
-        else:
-            outfile = open(output, 'wb')
+        tree, branch = bzrdir.BzrDir.open_containing_tree_or_branch(from_)[:2]
         # we may need to write data into branch's repository to calculate
         # the data to send.
         branch.lock_write()
@@ -4959,21 +4953,20 @@
                     public_branch=public_branch, patch_type=patch_type,
                     message=message)
 
-            outfile.writelines(directive.to_lines())
             if output is None:
-                subject = '[MERGE] '
-                if message is not None:
-                    subject += message
+                directive.compose_merge_request(mail_client, mail_to, body,
+                                                branch, tree)
+            else:
+                if output == '-':
+                    outfile = self.outf
                 else:
-                    revision = branch.repository.get_revision(revision_id)
-                    subject += revision.get_summary()
-                basename = directive.get_disk_name(branch)
-                mail_client.compose_merge_request(mail_to, subject,
-                                                  outfile.getvalue(),
-                                                  basename, body)
+                    outfile = open(output, 'wb')
+                try:
+                    outfile.writelines(directive.to_lines())
+                finally:
+                    if outfile is not self.outf:
+                        outfile.close()
         finally:
-            if output != '-':
-                outfile.close()
             branch.unlock()
 
 

=== modified file 'bzrlib/hooks.py'
--- bzrlib/hooks.py	2009-05-08 21:26:50 +0000
+++ bzrlib/hooks.py	2009-05-11 18:35:20 +0000
@@ -57,8 +57,8 @@
     ('bzrlib.version_info_formats.format_rio', 'RioVersionInfoBuilder.hooks'),
     'bzrlib.version_info_formats.format_rio', 'RioVersionInfoBuilderHooks')
 known_hooks.register_lazy(
-    ('bzrlib.mail_client', 'MailClient.hooks'),
-    'bzrlib.mail_client', 'MailClientHooks')
+    ('bzrlib.merge_directive', '_BaseMergeDirective.hooks'),
+    'bzrlib.merge_directive', 'MergeDirectiveHooks')
 
 
 def known_hooks_key_to_object((module_name, member_name)):

=== modified file 'bzrlib/mail_client.py'
--- bzrlib/mail_client.py	2009-05-08 21:26:50 +0000
+++ bzrlib/mail_client.py	2009-05-11 18:35:20 +0000
@@ -25,7 +25,6 @@
 from bzrlib import (
     email_message,
     errors,
-    hooks,
     msgeditor,
     osutils,
     urlutils,
@@ -35,22 +34,9 @@
 mail_client_registry = registry.Registry()
 
 
-class MailClientHooks(hooks.Hooks):
-
-    def __init__(self):
-        hooks.Hooks.__init__(self)
-        self.create_hook(hooks.HookPoint('merge_request_body',
-            "Called with a MergeRequestBodyParams when a body is needed for"
-            " a merge request.  Callbacks must return a body.  If more"
-            " than one callback is registered, the output of one callback is"
-            " provided to the next.", (1, 15, 0), False))
-
-
 class MailClient(object):
     """A mail client that can send messages with attachements."""
 
-    hooks = MailClientHooks()
-
     def __init__(self, config):
         self.config = config
 
@@ -86,13 +72,8 @@
         :param basename: The name to use for the attachment, e.g.
             "send-nick-3252"
         """
-        orig_body = body
         prompt = self._get_merge_prompt("Please describe these changes:", to,
                                         subject, directive)
-        for hook in self.hooks['merge_request_body']:
-            params = MergeRequestBodyParams(body, orig_body, directive,
-                                            to, basename, subject)
-            body = hook(params)
         self.compose(prompt, to, subject, directive,
             'x-patch', '.patch', basename, body)
 
@@ -107,17 +88,6 @@
         return ''
 
 
-class MergeRequestBodyParams(object):
-
-    def __init__(self, body, orig_body, directive, to, basename, subject):
-        self.body = body
-        self.orig_body = orig_body
-        self.directive = directive
-        self.to = to
-        self.basename = basename
-        self.subject = subject
-
-
 class Editor(MailClient):
     """DIY mail client that uses commit message editor"""
 

=== modified file 'bzrlib/merge_directive.py'
--- bzrlib/merge_directive.py	2009-05-06 05:36:28 +0000
+++ bzrlib/merge_directive.py	2009-05-11 18:35:20 +0000
@@ -23,6 +23,7 @@
     diff,
     errors,
     gpg,
+    hooks,
     registry,
     revision as _mod_revision,
     rio,
@@ -35,8 +36,35 @@
 from bzrlib.email_message import EmailMessage
 
 
+class MergeRequestBodyParams(object):
+
+    def __init__(self, body, orig_body, directive, to, basename, subject,
+                 branch, tree=None):
+        self.body = body
+        self.orig_body = orig_body
+        self.directive = directive
+        self.branch = branch
+        self.tree = tree
+        self.to = to
+        self.basename = basename
+        self.subject = subject
+
+
+class MergeDirectiveHooks(hooks.Hooks):
+
+    def __init__(self):
+        hooks.Hooks.__init__(self)
+        self.create_hook(hooks.HookPoint('merge_request_body',
+            "Called with a MergeRequestBodyParams when a body is needed for"
+            " a merge request.  Callbacks must return a body.  If more"
+            " than one callback is registered, the output of one callback is"
+            " provided to the next.", (1, 15, 0), False))
+
+
 class _BaseMergeDirective(object):
 
+    hooks = MergeDirectiveHooks()
+
     def __init__(self, revision_id, testament_sha1, time, timezone,
                  target_branch, patch=None, source_branch=None, message=None,
                  bundle=None):
@@ -240,6 +268,24 @@
                 target_repo.fetch(source_branch.repository, self.revision_id)
         return self.revision_id
 
+    def compose_merge_request(self, mail_client, to, body, branch, tree=None):
+        basename = self.get_disk_name(branch)
+        subject = '[MERGE] '
+        if self.message is not None:
+            subject += self.message
+        else:
+            revision = branch.repository.get_revision(self.revision_id)
+            subject += revision.get_summary()
+        orig_body = body
+        for hook in self.hooks['merge_request_body']:
+            params = MergeRequestBodyParams(body, orig_body, self,
+                                            to, basename, subject, branch,
+                                            tree)
+            body = hook(params)
+        mail_client.compose_merge_request(to, subject,
+                                          ''.join(self.to_lines()),
+                                          basename, body)
+
 
 class MergeDirective(_BaseMergeDirective):
 

=== modified file 'bzrlib/tests/test_mail_client.py'
--- bzrlib/tests/test_mail_client.py	2009-05-08 21:26:50 +0000
+++ bzrlib/tests/test_mail_client.py	2009-05-11 18:35:20 +0000
@@ -262,61 +262,3 @@
         self.assertEqual(dummy_client.args, (to, subject, directive))
         self.assertEqual(dummy_client.kwargs,
                          {"basename": basename, 'body': None})
-
-
-class TestHook(object):
-    def __init__(self, result=None):
-        self.calls = []
-        self.result = result
-
-    def __call__(self, params):
-        self.calls.append(params)
-        return self.result
-
-
-class HookMailClient(mail_client.MailClient):
-
-    def __init__(self, config):
-        self.body = None
-        self.config = config
-
-    def compose(self, prompt, to, subject, attachment, mime_subtype,
-                extension, basename=None, body=None):
-        self.body = body
-
-
-class TestBodyHook(tests.TestCase):
-
-    def compose_with_hooks(self, test_hooks):
-        for test_hook in test_hooks:
-            mail_client.MailClient.hooks.install_named_hook(
-                'merge_request_body', test_hook, 'test')
-            client = HookMailClient({})
-            client.compose_merge_request('jrandom@example.com',
-                'This code rox', {}, 'basename')
-        return client
-
-    def test_body_hook(self):
-        test_hook = TestHook('foo')
-        client = self.compose_with_hooks([test_hook])
-        self.assertEqual(1, len(test_hook.calls))
-        self.assertEqual('foo', client.body)
-        params = test_hook.calls[0]
-        self.assertIsInstance(params,
-                              mail_client.MergeRequestBodyParams)
-        self.assertIs(None, params.body)
-        self.assertIs(None, params.orig_body)
-        self.assertEqual('jrandom@example.com', params.to)
-        self.assertEqual('This code rox', params.subject)
-        self.assertEqual({}, params.directive)
-        self.assertEqual('basename', params.basename)
-
-    def test_body_hook_chaining(self):
-        test_hook1 = TestHook('foo')
-        test_hook2 = TestHook('bar')
-        client = self.compose_with_hooks([test_hook1, test_hook2])
-        self.assertEqual(None, test_hook1.calls[0].body)
-        self.assertEqual(None, test_hook1.calls[0].orig_body)
-        self.assertEqual('foo', test_hook2.calls[0].body)
-        self.assertEqual(None, test_hook2.calls[0].orig_body)
-        self.assertEqual('bar', client.body)

=== modified file 'bzrlib/tests/test_merge_directive.py'
--- bzrlib/tests/test_merge_directive.py	2009-03-23 14:59:43 +0000
+++ bzrlib/tests/test_merge_directive.py	2009-05-11 18:35:20 +0000
@@ -19,6 +19,7 @@
 from bzrlib import (
     errors,
     gpg,
+    mail_client,
     merge_directive,
     tests,
     )
@@ -697,3 +698,71 @@
         self.assertEqual('booga', md.patch)
         self.assertEqual('diff', md.patch_type)
         self.assertEqual('Hi mom!', md.message)
+
+
+class TestHook(object):
+    """Hook callback for test purposes."""
+
+    def __init__(self, result=None):
+        self.calls = []
+        self.result = result
+
+    def __call__(self, params):
+        self.calls.append(params)
+        return self.result
+
+
+class HookMailClient(mail_client.MailClient):
+    """Mail client for testing hooks."""
+
+    def __init__(self, config):
+        self.body = None
+        self.config = config
+
+    def compose(self, prompt, to, subject, attachment, mime_subtype,
+                extension, basename=None, body=None):
+        self.body = body
+
+
+class TestBodyHook(tests.TestCaseWithTransport):
+
+    def compose_with_hooks(self, test_hooks):
+        client = HookMailClient({})
+        for test_hook in test_hooks:
+            merge_directive.MergeDirective.hooks.install_named_hook(
+                'merge_request_body', test_hook, 'test')
+        tree = self.make_branch_and_tree('foo')
+        tree.commit('foo')
+        directive = merge_directive.MergeDirective2(
+            tree.branch.last_revision(), 'sha', 0, 0, 'sha',
+            source_branch=tree.branch.base,
+            base_revision_id=tree.branch.last_revision(),
+            message='This code rox')
+        directive.compose_merge_request(client, 'jrandom@example.com',
+            None, tree.branch)
+        return client, directive
+
+    def test_body_hook(self):
+        test_hook = TestHook('foo')
+        client, directive = self.compose_with_hooks([test_hook])
+        self.assertEqual(1, len(test_hook.calls))
+        self.assertEqual('foo', client.body)
+        params = test_hook.calls[0]
+        self.assertIsInstance(params,
+                              merge_directive.MergeRequestBodyParams)
+        self.assertIs(None, params.body)
+        self.assertIs(None, params.orig_body)
+        self.assertEqual('jrandom@example.com', params.to)
+        self.assertEqual('[MERGE] This code rox', params.subject)
+        self.assertEqual(directive, params.directive)
+        self.assertEqual('foo-1', params.basename)
+
+    def test_body_hook_chaining(self):
+        test_hook1 = TestHook('foo')
+        test_hook2 = TestHook('bar')
+        client = self.compose_with_hooks([test_hook1, test_hook2])[0]
+        self.assertEqual(None, test_hook1.calls[0].body)
+        self.assertEqual(None, test_hook1.calls[0].orig_body)
+        self.assertEqual('foo', test_hook2.calls[0].body)
+        self.assertEqual(None, test_hook2.calls[0].orig_body)
+        self.assertEqual('bar', client.body)

